/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <unistd.h>
#include <bundle.h>
#include <aul.h>
#include <dlog.h>
#include <vconf-internal-keys.h>
#include <app_common.h>
#include <Ecore.h>
#include <appcore_base.h>
#include <service_app.h>

#include "service_app_extension.h"
#include "service_app_internal.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "CAPI_APPFW_APPLICATION"

extern int app_control_create_event(bundle *data, struct app_control_s **app_control);

typedef enum {
	APP_STATE_NOT_RUNNING,
	APP_STATE_CREATING,
	APP_STATE_RUNNING,
} app_state_e;

struct app_event_handler {
	app_event_type_e type;
	app_event_cb cb;
	void *data;
	void* raw;
};

struct app_event_info {
	app_event_type_e type;
	void *value;
};

struct service_app_context {
	service_app_lifecycle_callback_s callback;
	void *data;
};

static struct service_app_context __context;
static app_state_e __app_state = APP_STATE_NOT_RUNNING;

static int __app_event_converter[APPCORE_BASE_EVENT_MAX] = {
	[APP_EVENT_LOW_MEMORY] = APPCORE_BASE_EVENT_LOW_MEMORY,
	[APP_EVENT_LOW_BATTERY] = APPCORE_BASE_EVENT_LOW_BATTERY,
	[APP_EVENT_LANGUAGE_CHANGED] = APPCORE_BASE_EVENT_LANG_CHANGE,
	[APP_EVENT_DEVICE_ORIENTATION_CHANGED] = APPCORE_BASE_EVENT_DEVICE_ORIENTATION_CHANGED,
	[APP_EVENT_REGION_FORMAT_CHANGED] = APPCORE_BASE_EVENT_REGION_CHANGE,
	[APP_EVENT_SUSPENDED_STATE_CHANGED] = APPCORE_BASE_EVENT_SUSPENDED_STATE_CHANGE,
};

static int __on_error(app_error_e error, const char *function, const char *description);

static int __service_app_create(void *data)
{
	appcore_base_on_create();

	if (__context.callback.create == NULL ||
			__context.callback.create(__context.data) == false)
		return __on_error(APP_ERROR_INVALID_CONTEXT, __FUNCTION__, "service_app_create_cb() returns false");

	return APP_ERROR_NONE;
}

static int __service_app_terminate(void *data)
{
	if (__context.callback.terminate)
		__context.callback.terminate(__context.data);

	appcore_base_on_terminate();

	return APP_ERROR_NONE;
}

static int __service_app_control(bundle *b, void *data)
{
	app_control_h app_control = NULL;

	LOGD("[SERVICE_APP] app_control callback");
	appcore_base_on_control(b);

	if (app_control_create_event(b, &app_control) != 0)
		return -1;

	if (__context.callback.app_control)
		__context.callback.app_control(app_control, __context.data);

	app_control_destroy(app_control);

	return 0;
}

static void __loop_init(int argc, char **argv, void *data)
{
	ecore_init();
}

static void __loop_fini(void)
{
	ecore_shutdown();
}

static void __loop_run(void *data)
{
	ecore_main_loop_begin();
}

static void __loop_exit(void *data)
{
	ecore_main_loop_quit();
}

static const char *__error_to_string(app_error_e error)
{
	switch (error) {
	case APP_ERROR_NONE:
		return "NONE";
	case APP_ERROR_INVALID_PARAMETER:
		return "INVALID_PARAMETER";
	case APP_ERROR_OUT_OF_MEMORY:
		return "OUT_OF_MEMORY";
	case APP_ERROR_INVALID_CONTEXT:
		return "INVALID_CONTEXT";
	case APP_ERROR_NO_SUCH_FILE:
		return "NO_SUCH_FILE";
	case APP_ERROR_ALREADY_RUNNING:
		return "ALREADY_RUNNING";
	default:
		return "UNKNOWN";
	}
}

static int __on_error(app_error_e error, const char *function, const char *description)
{
	if (description)
		LOGE("[%s] %s(0x%08x) : %s", function, __error_to_string(error), error, description);
	else
		LOGE("[%s] %s(0x%08x)", function, __error_to_string(error), error);

	return error;
}

EXPORT_API int service_app_main_ext(int argc, char **argv, service_app_lifecycle_callback_s *callback,
		service_app_loop_method_s *method, void *user_data)
{
	int ret;
	appcore_base_ops ops = appcore_base_get_default_ops();

	if (argc < 1 || argv == NULL || callback == NULL || method == NULL)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, NULL);

	if (callback->create == NULL)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, "service_app_create_cb() callback must be registered");

	if (__app_state != APP_STATE_NOT_RUNNING)
		return __on_error(APP_ERROR_ALREADY_RUNNING, __FUNCTION__, NULL);

	/* override methods */
	ops.create = __service_app_create;
	ops.terminate = __service_app_terminate;
	ops.control = __service_app_control;
	ops.run = method->run;
	ops.exit = method->exit;
	ops.init = method->init;
	ops.finish = method->fini;

	__context.callback = *callback;
	__context.data = user_data;

	__app_state = APP_STATE_CREATING;
	ret = appcore_base_init(ops, argc, argv, NULL);
	if (ret < 0) {
		__app_state = APP_STATE_NOT_RUNNING;
		return __on_error(APP_ERROR_INVALID_CONTEXT, __FUNCTION__, NULL);
	}

	appcore_base_fini();
	__app_state = APP_STATE_NOT_RUNNING;
	return APP_ERROR_NONE;
}

EXPORT_API int service_app_main(int argc, char **argv, service_app_lifecycle_callback_s *callback, void *user_data)
{
	service_app_loop_method_s method = {
		.init = __loop_init,
		.fini = __loop_fini,
		.run = __loop_run,
		.exit = __loop_exit,
	};

	return service_app_main_ext(argc, argv, callback, &method, user_data);
}

EXPORT_API void service_app_exit(void)
{
	appcore_base_exit();
}

static int __event_cb(void *event, void *data)
{
	app_event_handler_h handler = data;

	struct app_event_info app_event;

	app_event.type = handler->type;
	app_event.value = event;

	if (handler->cb)
		handler->cb(&app_event, handler->data);
	return 0;
}

EXPORT_API int service_app_add_event_handler(app_event_handler_h *event_handler, app_event_type_e event_type, app_event_cb callback, void *user_data)
{
	app_event_handler_h handler;

	if (event_handler == NULL || callback == NULL)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, "null parameter");

	if (event_type < APP_EVENT_LOW_MEMORY || event_type > APP_EVENT_SUSPENDED_STATE_CHANGED ||
			event_type == APP_EVENT_DEVICE_ORIENTATION_CHANGED)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, "invalid event type");

	handler = calloc(1, sizeof(struct app_event_handler));
	if (!handler)
		return __on_error(APP_ERROR_OUT_OF_MEMORY, __FUNCTION__, "insufficient memory");

	handler->type = event_type;
	handler->cb = callback;
	handler->data = user_data;
	handler->raw = appcore_base_add_event(__app_event_converter[event_type], __event_cb, handler);

	*event_handler = handler;

	return APP_ERROR_NONE;
}

EXPORT_API int service_app_remove_event_handler(app_event_handler_h event_handler)
{
	int ret;
	app_event_type_e type;

	if (event_handler == NULL)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, NULL);

	type = event_handler->type;
	if (type < APP_EVENT_LOW_MEMORY ||
			type > APP_EVENT_SUSPENDED_STATE_CHANGED ||
			type == APP_EVENT_DEVICE_ORIENTATION_CHANGED)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, NULL);

	ret = appcore_base_remove_event(event_handler->raw);
	free(event_handler);

	if (ret < 0)
		return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, NULL);

	return APP_ERROR_NONE;
}

EXPORT_API void service_app_exit_without_restart(void)
{
	aul_status_update(STATUS_NORESTART);
	appcore_base_exit();
}


